use core::mem;

use cortex_m::{delay::Delay, prelude::_embedded_hal_spi_FullDuplex};
use embedded_hal::digital::v2::OutputPin;
use fugit::{HertzU32, MegahertzU32};
use rp_pico::{
    hal::{
        gpio::{
            bank0::{Gpio2, Gpio3, Gpio4, Gpio5},
            Function, Output, Pin, PushPull, Spi,
        },
        spi::{self, Enabled, Spi as SpiDevice},
    },
    pac::{RESETS, SPI0},
};

pub struct SpiPins {
    pub sclk: Pin<Gpio2, Function<Spi>>,
    pub mosi: Pin<Gpio3, Function<Spi>>,
    pub miso: Pin<Gpio4, Function<Spi>>,
    pub cs: Pin<Gpio5, Output<PushPull>>,
}

pub struct SpiMouseSensor<L>
where
    L: FnMut(&str),
{
    pins: SpiPins,

    spi: SpiDevice<Enabled, SPI0, 8>,
    delay: Delay,

    pub log: L,
}

impl<L> SpiMouseSensor<L>
where
    L: FnMut(&str),
{
    const READ: u8 = 0b0111_1111;
    const WRITE: u8 = 0b1000_0000;

    const REG_PRODUCT_ID1: u8 = 0x00;
    const REG_PRODUCT_ID2: u8 = 0x01;
    const REG_MOTION_STATUS: u8 = 0x02;
    const REG_DELTA_X: u8 = 0x03;
    const REG_DELTA_Y: u8 = 0x04;
    const REG_OPERATION_MODE: u8 = 0x05;
    const REG_CONFIGURATION: u8 = 0x06;
    const REG_WRITE_PROTECT: u8 = 0x09;
    const REG_SLEEP1: u8 = 0x0A;
    const REG_SLEEP2: u8 = 0x0B;
    const REG_SLEEP3: u8 = 0x0C;
    const REG_CPI_X: u8 = 0x0D;
    const REG_CPI_Y: u8 = 0x0E;
    const REG_DELTA_XY_HI: u8 = 0x12;
    const REG_IQC: u8 = 0x13;
    const REG_SHUTTER: u8 = 0x14;
    const REG_FRAME_AVG: u8 = 0x17;
    const REG_MOUSE_OPTION: u8 = 0x19;
    const REG_SPI_MODE: u8 = 0x26;

    pub fn new(
        mut pins: SpiPins,
        resets: &mut RESETS,
        spio: SPI0,
        peripheral_clock: impl Into<HertzU32>,
        mut delay: Delay,
        mut log: L,
    ) -> Self {
        // Create an SPI driver instance for the SPI0 device
        let spi = spi::Spi::<_, _, 8>::new(spio);

        // Exchange the uninitialised SPI driver for an initialised one
        let spi = spi.init(
            resets,
            peripheral_clock,
            MegahertzU32::from_raw(1).convert(),
            &embedded_hal::spi::MODE_2,
        );

        if let Err(_err) = pins.cs.set_low() {
            log("cs high failed");
        }

        delay.delay_ms(2);

        if let Err(_err) = pins.cs.set_high() {
            log("cs high failed");
        }

        Self {
            pins,

            spi,
            delay,

            log,
        }
    }

    fn access<F, T>(&mut self, f: F) -> T
    where
        F: Fn(&mut Self) -> T,
    {
        if self.pins.cs.set_low().is_err() {
            (self.log)("cs low failed");
        }

        self.delay.delay_us(1);

        let res = f(self);

        if self.pins.cs.set_high().is_err() {
            (self.log)("cs low failed");
        }

        self.delay.delay_us(2);

        res
    }

    fn read(&mut self, address: u8) -> u8 {
        loop {
            let res = self.access(|me| {
                if me.spi.send(address & Self::READ).is_err() {
                    (me.log)("send address (read) failed");
                }

                if me.spi.send(0).is_err() {
                    (me.log)("send dummy (read) failed");
                }

                match me.spi.read() {
                    Ok(r) => r << 1,
                    Err(_) => {
                        (me.log)("read failed");
                        0
                    }
                }
            });

            if res != address {
                return res;
            }
        }
    }

    fn write(&mut self, address: u8, value: u8) {
        self.access(|me| {
            if me.spi.send(address | Self::WRITE).is_err() {
                (me.log)("send address (write) failed");
            }

            if me.spi.send(value).is_err() {
                (me.log)("send value (write) failed");
            }
        });
    }
}

impl<L> SpiMouseSensor<L>
where
    L: FnMut(&str),
{
    pub fn delta_x(&mut self) -> i8 {
        unsafe { mem::transmute(self.read(Self::REG_DELTA_X)) }
    }

    pub fn delta_y(&mut self) -> i8 {
        unsafe { mem::transmute(self.read(Self::REG_DELTA_Y)) }
    }

    pub fn product_id_1(&mut self) -> u8 {
        self.read(Self::REG_PRODUCT_ID1)
    }

    pub fn product_id_2(&mut self) -> u8 {
        self.read(Self::REG_PRODUCT_ID2)
    }

    pub fn operation_mode(&mut self) -> u8 {
        self.read(Self::REG_OPERATION_MODE)
    }

    pub fn motion_status(&mut self) -> u8 {
        self.read(Self::REG_MOTION_STATUS)
    }
}
